/**
 * $Revision: 3023 $
 * $Date: 2005-11-02 18:00:15 -0300 (Wed, 02 Nov 2005) $
 *
 * Copyright (C) 2005 Jive Software. All rights reserved.
 *
 * This software is published under the terms of the GNU Public License (GPL),
 * a copy of which is included in this distribution.
 */

package org.jivesoftware.smackx.commands;

import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.SmackConfiguration;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smackx.Form;
import org.jivesoftware.smackx.packet.AdHocCommandData;

/**
 * Represents a command that is in a remote location. Invoking one of the
 * execute, next, prev, cancel or complete methods results in executing that
 * action in the remote location. In response to that action the internal state
 * of the this command instance will change. For example, if the command is a
 * single stage command, then invoking the execute action will execute this
 * action in the remote location. After that the local instance will have a
 * state of "completed" and a form or notes that applies.
 * 
 * @author Gabriel Guardincerri
 * 
 */
public class RemoteCommand extends AdHocCommand {

    /**
     * The connection that is used to execute this command
     */
    private XMPPConnection connection;

    /**
     * The full JID of the command host
     */
    private String jid;

    /**
     * The session ID of this execution.
     */
    private String sessionID;

    /**
     * Creates a new RemoteCommand that uses an specific connection to execute a
     * command identified by <code>node</code> in the host identified by
     * <code>jid</code>
     * 
     * @param connection
     *            the connection to use for the execution.
     * @param node
     *            the identifier of the command.
     * @param jid
     *            the JID of the host.
     */
    protected RemoteCommand(XMPPConnection connection, String node, String jid) {
        super();
        this.connection = connection;
        this.jid = jid;
        this.setNode(node);
    }

    @Override
    public void cancel() throws XMPPException {
        executeAction(Action.cancel);
    }

    @Override
    public void complete(Form form) throws XMPPException {
        executeAction(Action.complete, form);
    }

    @Override
    public void execute() throws XMPPException {
        executeAction(Action.execute);
    }

    /**
     * Executes the default action of the command with the information provided
     * in the Form. This form must be the anwser form of the previous stage. If
     * there is a problem executing the command it throws an XMPPException.
     * 
     * @param response
     *            the form anwser of the previous stage.
     * @throws XMPPException
     */
    public void execute(Form form) throws XMPPException {
        executeAction(Action.execute, form);
    }

    @Override
    public void next(Form form) throws XMPPException {
        executeAction(Action.next, form);
    }

    @Override
    public void prev() throws XMPPException {
        executeAction(Action.prev);
    }

    private void executeAction(Action action) throws XMPPException {
        executeAction(action, null);
    }

    /**
     * Executes the <code>action</codo> with the <code>form</code>.
     * The action could be any of the available actions. The form must
     * be the anwser of the previous stage. It can be null if it is the first stage.
     * @param action the action to execute
     * @param form the form with the information
     * @throws XMPPException if there is a problem executing the command.
     */
    private void executeAction(Action action, Form form) throws XMPPException {
        // TODO Check that all the requiered fields of the form were filled, if
        // not throw the corresponding exeption. This will make a faster response,
        // since the request is stoped before it's sended.
        AdHocCommandData data = new AdHocCommandData();
        data.setType(IQ.Type.SET);
        data.setTo(getOwnerJID());
        data.setNode(getNode());
        data.setSessionID(sessionID);
        data.setAction(action);

        if (form != null) {
            data.setForm(form.getDataFormToSend());
        }

        PacketCollector collector = connection
                .createPacketCollector(new PacketIDFilter(data.getPacketID()));

        connection.sendPacket(data);

        Packet response = collector.nextResult(SmackConfiguration
                .getPacketReplyTimeout());

        // Cancel the collector.
        collector.cancel();
        if (response == null) {
            throw new XMPPException("No response from server on status set.");
        }
        if (response.getError() != null) {
            throw new XMPPException(response.getError());
        }

        AdHocCommandData responseData = (AdHocCommandData) response;
        this.sessionID = responseData.getSessionID();
        super.setData(responseData);
    }

    @Override
    public String getOwnerJID() {
        return jid;
    }
}
